from datetime import time
from django.contrib.sites.managers import CurrentSiteManager
from django.contrib.sites.models import Site
from django.db import models
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db.models.deletion import CASCADE
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from user_messages import api as user_messages
from user_messages.models import Message


def set_default_site_id():
    return Site.objects.get_current()

def set_default_message():
    default = {
        "title": "",
        "description": "",
        "accept_button": "Accept",
        "accept_url": "user_notifications:notification-confirm",
        "decline_button": "Decline",
        "decline_url": "user_notifications:notification-decline"
    }
    return default

class DisplayType(models.IntegerChoices):
    BOOTSTRAP_ALERT = 100, _("Bootstrap Alert")
    MODAL_INFO = 200, _("Modal Info")
    MODAL_FORM = 201, _("Modal Form")


class Notification(models.Model):
    """
    Class that helps creating site wide notifications.
    Notifications can be displayed as a Bootstrap Alert or as a Bootstrap Modal.
    One can choose to show the message to the user once or on every view until the user takes action.
    One can choose to show the message from or to a certain date with start and end date.
    One can choose to add additional rules implemented by the developer by using the RuleBase Class
    The notification has a JSON filed called message with the following default attributes:
        title: <Title for the modal (Ignored by the bootstrap Alert)>
        description: <The modal body or alert description>
        accept_button: <Text inside the accept button in the modal (Ignored by the bootstrap Alert)>
        accept_url: <POST path for a user if they decide to accept the notification (Ignored by the bootstrap Alert)>
        decline_button: <Text inside the decline button in the modal (Ignored by the bootstrap Alert)>
        accept_url: <POST path for a user if they decide to decline the notification (Ignored by the bootstrap Alert)>
    """
    name = models.CharField(verbose_name=_("Notification Name"), blank=False, null=False, max_length=65)
    active = models.BooleanField(verbose_name=_("Active"), default=False)
    sites = models.ManyToManyField(Site, verbose_name=_("Sites"), blank=True, related_name="notifications")
    display_type = models.IntegerField(verbose_name=_("Message Type"), choices=DisplayType.choices, default=DisplayType.BOOTSTRAP_ALERT)
    message = models.JSONField(verbose_name=_("Message Content"), blank=True, null=True, default=set_default_message)
    deliver_once = models.BooleanField(verbose_name=_("Deliver Once"), default=True)
    start_date = models.DateTimeField(_("Start Date"), blank=True, null=True, help_text=_("Notification Start Date?"))
    end_date = models.DateTimeField(_("End Date"), blank=True, null=True, help_text=_("Notification End Date?"))
    rules = models.JSONField(default=list, blank=True, null=True)
    meta = models.JSONField(default=dict, blank=True, null=True)

    objects = models.Manager()
    on_site = CurrentSiteManager()

    def __str__(self):
        return self.name

    def add_user_message(self, user):
        self.message['display_type'] = self.display_type
        self.message['pk'] = self.pk
        user_messages.info(user, self.name, deliver_once=self.deliver_once, meta=self.message)
    
    def already_exists(self, user):
        if Message.objects.filter(user=user, message=self.name).count():
            return True
        return False

    def is_between_dates(self):
        if not self.start_date and not self.end_date:
            return True
        elif (not self.start_date or timezone.now() > self.start_date) and (not self.end_date or timezone.now() < self.end_date):
            return True
        return False

    def save_accepted(self, user, site):
        if 'accepted' not in self.meta.keys():
            self.meta['accepted'] = dict()

        if site.domain not in self.meta['accepted'].keys():
            self.meta['accepted'][site.domain] = dict()

        if user.username not in self.meta['accepted'][site.domain].keys():
            self.meta['accepted'][site.domain][user.username] = list()

        self.meta['accepted'][site.domain][user.username].append(f"{timezone.now():%Y-%m-%d %H:%M:%S}")
        self.save()

    def save_declined(self, user, site):
        if 'declined' not in self.meta.keys():
            self.meta['declined'] = dict()

        if site.domain not in self.meta['declined'].keys():
            self.meta['declined'][site.domain] = dict()

        if user.username not in self.meta['declined'][site.domain].keys():
            self.meta['declined'][site.domain][user.username] = list()

        self.meta['declined'][site.domain][user.username].append(f"{timezone.now():%Y-%m-%d %H:%M:%S}")
        self.save()

    def save_user_acknowledgement(self, user, site, accepted):
        if self.deliver_once:
            return None     # Notifications should be deliver_once=False to work as an ack
        try:
            user_message = Message.objects.get(user=user, message=self.name, deliver_once=False, delivered_at__isnull=True)
        except ObjectDoesNotExist as error:
            return None
        except MultipleObjectsReturned as error:
            user_message = Message.objects.filter(user=user, message=self.name, deliver_once=False, delivered_at__isnull=True).order_by('created_at').first()

        user_message.delivered_at = timezone.now()
        user_message.deliver_once = True
        user_message.save()
        if accepted:
            self.save_accepted(user, site)
        else:
            self.save_declined(user, site)




from django.contrib import admin
from user_notifications.models import Notification

# Register your models here.
###############
# MODEL ADMINS
###############
class NotificationAdmin(admin.ModelAdmin):
    fields = ('active', 'deliver_once', 'name', 'sites', 'display_type', 'message', 'rules', 'meta')
    list_display = ('pk', 'name', 'active', 'deliver_once', 'display_type')


admin.site.register(Notification, NotificationAdmin)


from user_notifications.rules import RuleBase


class NoRuleNameExample(RuleBase):

    def __init__(self, notification, user):
        super().__init__(notification, user)

class DoesNotApplyRuleExample(RuleBase):
    RULE_NAME = "DoesNotApplyRuleExample"


from django.utils import timezone
from django.db.models import Q
from user_messages import api as user_messages
from user_notifications.models import Notification


class RuleBase:
    RULE_NAME = None
    notification = None
    user = None
    apply_rule = False

    def __init__(self, notification, user):
        if not self.RULE_NAME:
            raise NotImplementedError
        if not type(notification) == Notification:
            raise TypeError("passed notification variable is not a Notification type")
        if not user:
            raise TypeError("user cannot be None type")

        self.notification = notification
        self.user = user
    
    def __str__(self):
        return self.RULE_NAME

    def does_rule_apply(self):
        raise NotImplementedError()

    def gt(self, value):
        raise NotImplementedError()

    def gte(self, value):
        raise NotImplementedError()

    def lt(self, value):
        raise NotImplementedError()

    def lte(self, value):
        raise NotImplementedError()



class RuleConstructorBase:
    """
    You can create you own rule contructor or implement another
    design to implement and process rules. This is just an example"""

    def create_rule(notification, rule_name, user):
        raise NotImplementedError()



    



from datetime import timedelta
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.utils import timezone
from django.test import TestCase, Client
from user_messages.models import Message
from user_notifications.models import Notification

User = get_user_model()

# Create your tests here.
class NotificationModelTests(TestCase):

    fixtures = ['user', 'unit_test']

    def setUp(self):
        self.client = Client()
        self.user = User.objects.get(pk=1)
        self.client.force_login(self.user)

    def test_queue_expire_after_date(self):
        notification = Notification.objects.get(pk=3)
        self.assertTrue(notification.is_between_dates)

    def test_dont_queue_expire_after_date(self):
        notification = Notification.objects.get(pk=3)
        notification.end_date = (timezone.now() - timedelta(days=1))
        self.assertFalse(notification.is_between_dates())

    def test_queue_start_after_date(self):
        notification = Notification.objects.get(pk=4)
        self.assertTrue(notification.is_between_dates())

    def test_dont_queue_start_after_date(self):
        notification = Notification.objects.get(pk=4)
        notification.start_date = (timezone.now() + timedelta(days=1))
        self.assertFalse(notification.is_between_dates())

    def test_queue_between_start_and_end_date(self):
        notification = Notification.objects.get(pk=5)
        self.assertTrue(notification.is_between_dates())
    
    def test_add_user_message_success(self):
        notification = Notification.objects.get(pk=5)
        self.assertFalse(Message.objects.all().count())
        notification.add_user_message(User.objects.get(pk=1))
        self.assertTrue(Message.objects.all().count())

    def test_already_exists_true(self):
        user = User.objects.get(pk=1)
        notification = Notification.objects.get(pk=5)
        self.assertFalse(Message.objects.all().count())
        notification.add_user_message(user)
        self.assertTrue(notification.already_exists(user))
    
    def test_already_exists_false(self):
        notification = Notification.objects.get(pk=5)
        self.assertFalse(notification.already_exists(User.objects.get(pk=1)))

    def test_save_accepted(self):
        user = User.objects.get(pk=1)
        notification = Notification.objects.get(pk=5)
        notification.save_accepted(user, Site.objects.get(pk=1))
        self.assertIn('accepted', notification.meta.keys())
        self.assertIn(user.username, notification.meta['accepted'][Site.objects.get(pk=1).domain].keys())

    def test_save_decline(self):
        user = User.objects.get(pk=1)
        notification = Notification.objects.get(pk=5)
        notification.save_declined(user, Site.objects.get(pk=1))
        self.assertIn('declined', notification.meta.keys())
        self.assertIn(user.username, notification.meta['declined'][Site.objects.get(pk=1).domain].keys())
        
    def test_save_user_acknowledgement(self):
        user = User.objects.get(pk=1)
        notification = Notification.objects.get(pk=2)
        notification.add_user_message(user)
        notification.refresh_from_db()
        notification.save_user_acknowledgement(user, Site.objects.get(pk=1), True)
        self.assertIn('accepted', notification.meta.keys())
        self.assertIn(user.username, notification.meta['accepted'][Site.objects.get(pk=1).domain].keys())

        




        



from django.test import TestCase
from django.contrib.auth import get_user_model
from user_notifications.models import Notification
from user_notifications.tests.example_rules import DoesNotApplyRuleExample, NoRuleNameExample


User = get_user_model()
    

class RuleBaseTests(TestCase):

    fixtures = ['user', 'unit_test']

    def test_rule_name_not_implemented_error(self):
        with self.assertRaises(NotImplementedError) as error:
            invalid_rule = NoRuleNameExample(None, None)

    def test_rule_notification_incorrect_type_implemented_error(self):
        with self.assertRaises(TypeError) as error:
            invalid_rule = DoesNotApplyRuleExample("str instead of notification", User.objects.get(pk=1))

    def test_does_rule_apply_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            invalid_rule = DoesNotApplyRuleExample(Notification.objects.all().first(), User.objects.get(pk=1))
            invalid_rule.does_rule_apply()
        



# Generated by Django 3.1.13 on 2021-11-16 18:06

import django.contrib.sites.managers
from django.db import migrations, models
import django.db.models.manager
import user_notifications.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=65, verbose_name='Notification Name')),
                ('active', models.BooleanField(default=False, verbose_name='Active')),
                ('display_type', models.IntegerField(choices=[(100, 'Bootstrap Alert'), (200, 'Modal')], default=100, verbose_name='Message Type')),
                ('message', models.JSONField(blank=True, default=user_notifications.models.set_default_message, null=True, verbose_name='Message Content')),
                ('deliver_once', models.BooleanField(default=True, verbose_name='Deliver Once')),
                ('start_date', models.DateTimeField(blank=True, help_text='Notification Start Date?', null=True, verbose_name='Start Date')),
                ('end_date', models.DateTimeField(blank=True, help_text='Notification End Date?', null=True, verbose_name='End Date')),
                ('rules', models.JSONField(blank=True, default=list, null=True)),
                ('meta', models.JSONField(blank=True, default=dict, null=True)),
                ('sites', models.ManyToManyField(blank=True, related_name='notifications', to='sites.Site', verbose_name='Sites')),
            ],
            managers=[
                ('objects', django.db.models.manager.Manager()),
                ('on_site', django.contrib.sites.managers.CurrentSiteManager()),
            ],
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-01 17:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user_notifications', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='display_type',
            field=models.IntegerField(choices=[(100, 'Bootstrap Alert'), (200, 'Modal Info'), (201, 'Modal Form')], default=100, verbose_name='Message Type'),
        ),
    ]



from django.contrib.sites.shortcuts import get_current_site

#############
# Requst and Session Utils
def get_site_from_request(request):
    if hasattr(request, 'site'):
        return request.site
    return get_current_site(request)